/*
 *  tools.c -- misc utils
 *
 *  tools.c is a part of binkd project
 *
 *  Copyright (C) 1996  Dima Maloff, 5047/13
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version. See COPYING.
 */

/*
 * $Id: tools.c,v 1.7 1996/12/29 09:47:42 mff Exp mff $
 *
 * $Log: tools.c,v $
 * Revision 1.7  1996/12/29  09:47:42  mff
 * Added: create_empty_sem_file()
 *
 * Revision 1.6  1996/12/14  07:13:51  mff
 * log() is now Log()
 *
 * Revision 1.5  1996/12/05  08:01:11  mff
 * Changed parse_args(), removed free_args()
 *
 * Revision 1.4  1996/12/05  07:24:35  mff
 * Format strings for Log() should have no "\n" at the end
 *
 * Revision 1.3  1996/12/05  04:30:14  mff
 * Now we don't mkdir drives
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#ifdef HAVE_UNISTD
#include <unistd.h>
#endif
#ifdef HAVE_IOH
#include <io.h>
#endif

#include "Config.h"
#include "sys.h"
#include "tools.h"

int mkpath0 (const char *path0)
{
  struct stat sb;

  if (stat (path0, &sb) == -1)
  {
    char *path, *s;
    int retval;

    if ((path = strdup (path0)) == 0)
      return -1;
    if ((s = max (strrchr (path, '\\'), strrchr (path, '/'))) != 0)
    {
      *(s++) = 0;
      if (mkpath0 (path) == -1)
      {
	free (path);
	return -1;
      }
    }
    else
      s = path;
    if (path0[strlen (path0) - 1] != ':')
      retval = MKDIR (path0);
    free (path);
    return retval;
  }
  else if (sb.st_mode & S_IFDIR)
    return 0;

  errno = ENOTDIR;
  return -1;
}

int mkpath (char *s)
{
  char path[MAXPATHLEN];

  strnzcpy (path, s, MAXPATHLEN);
  if ((s = max (strrchr (path, '\\'), strrchr (path, '/'))) == 0)
    return 0;
  *s = 0;
  return mkpath0 (path);
}

long rnd ()
{
  static int i;

  if (!i)
  {
    i = 1;
    srand (time (0));
  }
  return time (0) + rand ();
}

/*
 * 1 -- created, 0 -- already busy
 */
int create_empty_sem_file (char *name)
{
  int h;

  if ((h = open (name, O_RDWR | O_CREAT | O_EXCL, 0666)) == -1)
    return 0;
  close (h);
  return 1;
}

int create_sem_file (char *name)
{
  FILE *f;
  int h;

  if ((h = open (name, O_RDWR | O_CREAT | O_EXCL, 0666)) == -1)
    return 0;
  if((f = fdopen(h, "w")) == 0)
    return 0;
  fprintf (f, "%i", (int) getpid ());
  fclose (f);
  return 1;
}

void Log (int lev, char *s,...)
{
  extern char logpath[];
  extern int loglevel;
  static int first_time = 1;

  char timebuf[60];
  time_t t;
  struct tm *tm;
  va_list ap;
  static const char *marks = "!?+-";
  char ch = (lev < (int) strlen (marks)) ? marks[lev] : ' ';

  if (lev <= loglevel)
  {
    time (&t);
    tm = localtime (&t);
    strftime (timebuf, sizeof (timebuf), "%H:%M", tm);
    fprintf (stderr, "%c %s [%i] ", ch, timebuf, (int) PID ());
    va_start (ap, s);
    vfprintf (stderr, s, ap);
    va_end (ap);
    fputc ('\n', stderr);
    if (*logpath)
    {
      FILE *logfile = 0;
      int i;

      for (i = 0; logfile == 0 && i < 10; ++i)
      {
	logfile = fopen (logpath, "a");
      }
      if (logfile)
      {
	if (first_time)
	{
	  fputc ('\n', logfile);
	  first_time = 0;
	}
	strftime (timebuf, sizeof (timebuf), "%d %b %H:%M:%S", tm);
	fprintf (logfile, "%c %s [%i] ", ch, timebuf, (int) PID ());
	va_start (ap, s);
	vfprintf (logfile, s, ap);
	va_end (ap);
	fputc ('\n', logfile);
	fclose (logfile);
      }
      else
	fprintf (stderr, "Cannot open %s: %s!\n", logpath, strerror (errno));
    }
  }
  if (lev == 0)
    exit (1);
}

int memicmp (const void *s1, const void *s2, size_t n)
{
  int i;

  for (i = 0; i < (int) n; ++i)
    if (tolower (((char *) s1)[i]) != tolower (((char *) s2)[i]))
      return (tolower (((char *) s1)[i]) - tolower (((char *) s2)[i]));

  return 0;
}

int stricmp (const char *s1, const char *s2)
{
  int i;

  for (i = 0;; ++i)
  {
    if (tolower (s1[i]) != tolower (s2[i]))
      return (tolower (s1[i]) - tolower (s2[i]));
    if (!s1[i])
      return 0;
  }
}

int strnicmp (const char *s1, const char *s2, size_t n)
{
  int i;

  for (i = 0; i < (int) n; ++i)
  {
    if (tolower (s1[i]) != tolower (s2[i]))
      return (tolower (s1[i]) - tolower (s2[i]));
    if (!s1[i])
      return 0;
  }
  return 0;
}

/*
 * Quotes all control chars. free() it's retval!
 */
char *strquote (char *s)
{
  char *r = malloc (strlen (s) * 4 + 1);
  int i;

  assert (r != 0);
  for (i = 0; *s; ++s)
  {
    if (iscntrl (*s) || *(unsigned char *) s == 0xff)
    {
      sprintf (r + i, "\\x%02x", *(unsigned char *) s);
      i += 4;
    }
    else
      r[i++] = *s;
  }
  r[i] = 0;
  return r;
}

/*
 * Removes controls.
 */
char *strwipe (char *s)
{
  int i;

  for (i = 0; s[i]; ++i)
    if (iscntrl (s[i]) || ((unsigned char *) s)[i] == 0xff ||
	s[i] == '\\' || s[i] == '/' || s[i] == ' ')
      s[i] = '_';
  return s;
}

/*
 * Copyes not more than len chars from src into dst, but, unlike strncpy(),
 * it appends '\0' even if src is longer than len.
 */
char *strnzcpy (char *dst, const char *src, size_t len)
{
  dst[len - 1] = 0;
  return strncpy (dst, src, len - 1);
}

char *strnzcat (char *dst, const char *src, size_t len)
{
  int x = strlen (dst);

  return strnzcpy (dst + x, src, len - x);
}

/*
 * Src is a source string, n is a word number (1...), returned string must
 * be free()'d. If subst_vars == 1 env. vars will be substed for %VAR%
 */
char *getwordx (char *src, int n, int subst_vars)
{
  char *dest = strdup (src);
  char quoted;
  int i;

  while (1)
  {
    while (*src && isspace (*src))
      ++src;

    if (!*src)
    {
      free (dest);
      return 0;
    }

    if (*src == '\"')
    {
      quoted = 1;
      ++src;
    }
    else
      quoted = 0;

    for (i = 0;; ++i, ++src)
    {
      if (*src == '\\')
      {
	int base = 8, ch = 0;

	++src;
	if (!isdigit (*src) && *src != 'x' && *src != 'X')
	{
	  switch (*src)
	    {
	      case 'a':		    /* bell */
		ch = '\a';
		break;
	      case 'b':		    /* bksp */
		ch = '\b';
		break;
	      case 'f':		    /* ff */
		ch = '\f';
		break;
	      case 'n':		    /* lf */
		ch = '\n';
		break;
	      case 'r':		    /* cr */
		ch = '\r';
		break;
	      case 't':		    /* tab */
		ch = '\t';
		break;
	      case 'E':		    /* ESC */
		ch = 0033;
		break;
	      default:
		ch = *src;
		break;
	    }
	  ++src;
	}
	else
	{
	  if (toupper (*src) == 'X')
	  {
	    ++src;
	    base = 16;
	  }
	  for (; isxdigit (*src); ++src)
	  {
	    if (isdigit (*src))
	      ch = ch * base + *src - '0';
	    else if (base == 16)
	      ch = ch * base + tolower (*src) - 'a' + 0xa;
	    else
	      break;
	  }
	}
	dest[i] = (char) ch;
	--src;
      }
      else if (!*src || *src == '\n' || *src == '\r')
	break;
      else if (isspace (*src) && !quoted)
	break;
      else if (src[0] == '\"' && isspace (src[1]) && quoted)
      {
	++src;
	break;
      }
      else if (*src == '%' && subst_vars)
      {
	++src;
	if (*src == '%')
	{
	  dest[i] = '%';
	}
	else
	{
	  char varname[MAX_ENV_VAR_NAME + 1];
	  char *var;
	  unsigned k;

	  for (k = 0; k < sizeof (varname) &&
	       !isspace (*src) &&
	       *src != '%'; ++k)
	    varname[k] = toupper (src++[0]);
	  varname[k] = 0;

	  if ((var = getenv (varname)) == 0)
	    var = "";

	  dest = realloc (dest, strlen (src) + strlen (var) + 1);
	  assert (dest != 0);

	  strcpy (dest + i, var);
	  i += strlen (var) - 1;
	}
      }
      else
	dest[i] = *src;
    }

    if (--n == 0)
    {
      dest[i] = 0;
      return dest;
    }
  }
}

/*
 * Splits args ASCIIZ string onto argc separate words,
 * saves them as argv[0]...argv[argc-1]. Logs error
 * "ID: cannot parse args", if args containes less than argc words.
 */
int parse_args (int argc, char *argv[], char *src, char *ID)
{
  int i = 0;

  while (i < argc)
  {
    while (*src && isspace (*src))
      ++src;
    if (!*src)
      break;
    argv[i] = src;
    while (*src && !isspace (*src))
      ++src;
    ++i;
    if (!*src)
      break;
    src++[0] = 0;
  }
  if (i < argc)
  {
    Log (1, "%s: cannot parse args", ID, src);
    return 0;
  }
  else
    return 1;
}
